import importlib
import hashlib
import inspect

from falcon_core.config import settings


def import_object(m):
    m = m.split('.')
    return getattr(importlib.import_module('.'.join(m[:-1])), m[-1], None)


def load_middleware(m):
    m = import_object(m)
    if inspect.isclass(m):
        return m()
    return m


def flatten(l):
    f = []
    for i in l:
        if isinstance(i, (list, tuple)):
            f.extend(flatten(i))
        else:
            f.append(i)
    return f


def encrypt_sha1(text):
    return hashlib.sha1(str(text).encode()).hexdigest()


def encrypt_sha256(text):
    return hashlib.sha256(str(text).encode()).hexdigest()


def encrypt_sha1_with_secret_key(text):
    return encrypt_sha1(text + settings.SECRET_KEY)


def encrypt_sha256_with_secret_key(text):
    return encrypt_sha256(text + settings.SECRET_KEY)


"""
Generate dict from instance
"""


def register_rule(name, rule=None):
    if DictGenerator.rules is not None:
        if rule:
            DictGenerator.rules.register(name, rule)
        else:
            def wraps_rule(rule):
                DictGenerator.rules.register(name, rule)
                return rule

            return wraps_rule


def dict_from_obj(obj, data, iterable=False):
    dict_generator = DictGenerator(obj, data, iterable)
    return dict_generator.generate()


class Rule:
    generator = None

    def __init__(self, obj, data=None):
        self.obj = obj
        self.data = data

    def generate(self):
        raise NotImplementedError


class Rules(dict):
    generator = None

    def __set_name__(self, owner, name):
        self.generator = owner

    def register(self, name, rule):
        rule.generator = self.generator
        self.update(**{name: rule})


class DictGenerator:
    rules = Rules()

    def __init__(self, obj, data, iterable=False):
        self.obj = obj
        self.data = data
        self.iterable = iterable

    def generate(self):
        if self.iterable:
            return ObjectsRule(self.obj, self.data).generate()
        return ObjectRule(self.obj, self.data).generate()


@register_rule('string')
class StrRule(Rule):
    def generate(self):
        return str(self.obj)


@register_rule('integer')
class IntRule(Rule):
    def generate(self):
        return int(self.obj)


@register_rule('float')
class FloatRule(Rule):
    def generate(self):
        return float(self.obj)


@register_rule('boolean')
class FloatRule(Rule):
    def generate(self):
        return bool(self.obj)


@register_rule('object')
class ObjectRule(Rule):
    def generate(self):
        data = {}
        for field, rule, *d in self.data:
            field, name = field.split(':') if ':' in field else (field, field)
            data[name] = self.generator.rules[rule](getattr(self.obj, field), *d).generate()
        return data


@register_rule('objects')
class ObjectsRule(Rule):
    def generate(self):
        return [ObjectRule(obj_item, self.data).generate() for obj_item in self.obj]



from importlib import import_module

from falcon import API

from falcon_core.config import settings

from falcon_core.utils import load_middleware, flatten


def get_wsgi_application():

    middleware = [load_middleware(m) for m in settings.MIDDLEWARE]

    application = API(middleware=middleware)

    for k, v in settings.ROUTER_CONVERTERS:
        application.router_options.converters[k] = v

    for route in flatten(import_module(settings.ROUTES).routes):
        application.add_route(route.template_uri, route.resource, *route.args, **route.kwargs)

    return application



import json

import falcon


class Resource:
    use_token = False
    storage = {}

    def json_load_from_req_stream(self, req):
        try:
            if req.content_length:
                data = json.load(req.stream)
            else:
                data = {}
            return data
        except json.JSONDecodeError:
            raise falcon.HTTPBadRequest

    def on_get(self, req, resp, **kwargs):
        self.get(req, resp, **kwargs)

    def get(self, req, resp, **kwargs):
        resp.status = falcon.HTTP_NOT_FOUND

    def on_post(self, req, resp, **kwargs):
        self.post(req, resp, self.json_load_from_req_stream(req), **kwargs)

    def post(self, req, resp, data, **kwargs):
        resp.status = falcon.HTTP_NOT_FOUND

    def on_put(self, req, resp, **kwargs):
        self.put(req, resp, self.json_load_from_req_stream(req), **kwargs)

    def put(self, req, resp, data, **kwargs):
        resp.status = falcon.HTTP_NOT_FOUND

    def on_delete(self, req, resp, **kwargs):
        self.delete(req, resp, self.json_load_from_req_stream(req), **kwargs)

    def delete(self, req, resp, data, **kwargs):
        resp.status = falcon.HTTP_NOT_FOUND



from importlib import import_module

from falcon_core.utils import flatten


class RouteError(ValueError):
    pass


class Route:
    def __init__(self, template_uri, resource, *args, **kwargs):
        if template_uri.startswith('/') and template_uri.endswith('/'):
            self.template_uri = template_uri
            self.resource = resource
            self.args = args
            self.kwargs = kwargs
        else:
            raise RouteError(f'Route template_uri \'{template_uri}\' is not valid.')


def route(*args, **kwargs):
    ri = Route(*args, **kwargs)
    if isinstance(ri.resource, list):
        ri.resource = flatten(ri.resource)
        for r in ri.resource:
            r.template_uri = f'{ri.template_uri[:-1]}{r.template_uri}'
        return ri.resource
    return ri


def include(routes):
    if isinstance(routes, str):
        return import_module(routes).routes
    return routes
